/* GStreamer Element
 * Copyright (C) 2006 Mark Nauwelaerts <mnauw@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1307, USA.
 */

/**
 * SECTION:element-tsdup
 *
 * <refsect2>
 * <para>
 * Duplicates PTS/DTS timestamps on buffers by copying one of PTS or DTS
 * to the other.  The precise duplication operation performed is determined
 * by property <link linkend="GstEntransTsDup--operation">operation</link> which sets
 * the destination of the duplication (the source being then obviously
 * the other one).
 * </para>
 * </refsect2>
 *
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "plugin-entrans.h"

#include <string.h>


#define GST_TYPE_TS_DUP \
  (gst_ts_dup_get_type())
#define GST_TS_DUP(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_TS_DUP,GstTsDup))
#define GST_TS_DUP_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_TS_DUP,GstTsDupClass))
#define GST_IS_TS_DUP(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_TS_DUP))
#define GST_IS_TS_DUP_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_TS_DUP))


typedef struct _GstTsDup GstTsDup;
typedef struct _GstTsDupClass GstTsDupClass;

struct _GstTsDup
{
  GstBaseTransform parent;

  /* properties */
  GstClockTime method;
};


struct _GstTsDupClass
{
  GstVideoFilterClass parent_class;
};

GST_DEBUG_CATEGORY_STATIC (shift_debug);
#define GST_CAT_DEFAULT shift_debug


/* signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_METHOD
      /* FILL ME */
};

#define DEFAULT_METHOD    METHOD_TO_INVALID

static GstStaticPadTemplate gst_ts_dup_src_template =
GST_STATIC_PAD_TEMPLATE (GST_BASE_TRANSFORM_SRC_NAME,
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

static GstStaticPadTemplate gst_ts_dup_sink_template =
GST_STATIC_PAD_TEMPLATE (GST_BASE_TRANSFORM_SINK_NAME,
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

static GstFlowReturn gst_ts_dup_transform_ip (GstBaseTransform * btrans,
    GstBuffer * in);
static gboolean gst_ts_dup_start (GstBaseTransform * btrans);
static gboolean gst_ts_dup_stop (GstBaseTransform * btrans);

static void gst_ts_dup_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_ts_dup_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

#define parent_class gst_ts_dup_parent_class
typedef GstTsDupClass GstEntransTsDupClass;
typedef GstTsDup GstEntransTsDup;
G_DEFINE_TYPE (GstEntransTsDup, gst_ts_dup, GST_TYPE_BASE_TRANSFORM);

#define GST_TYPE_TS_DUP_METHOD (gst_ts_dup_method_get_type ())

typedef enum
{
  METHOD_TO_INVALID,
  METHOD_TO_PTS,
  METHOD_TO_INVALID_PTS,
  METHOD_TO_DTS,
  METHOD_TO_INVALID_DTS
} GstTsDupMethods;

static GType
gst_ts_dup_method_get_type (void)
{
  static GType ts_dup_method_type = 0;

  if (!ts_dup_method_type) {
    static const GEnumValue ts_dup_methods[] = {
      {METHOD_TO_INVALID, "To invalid ts", "invalid"},
      {METHOD_TO_PTS, "To PTS", "pts"},
      {METHOD_TO_INVALID_PTS, "To invalid PTS", "invalid-pts"},
      {METHOD_TO_DTS, "To DTS", "dts"},
      {METHOD_TO_INVALID_DTS, "To invalid DTS", "invalid-dts"},
      {0, NULL, NULL},
    };

    ts_dup_method_type =
        g_enum_register_static ("GstTsDupMethods", ts_dup_methods);
  }

  return ts_dup_method_type;
}

static void
gst_ts_dup_class_init (GstTsDupClass * g_class)
{
  GObjectClass *gobject_class;
  GstElementClass *element_class;
  GstBaseTransformClass *trans_class;

  gobject_class = G_OBJECT_CLASS (g_class);
  element_class = GST_ELEMENT_CLASS (g_class);
  trans_class = GST_BASE_TRANSFORM_CLASS (g_class);

  GST_ENTRANS_DEBUG_CATEGORY_INIT (shift_debug, "tsdup", 0, "tsdup");

  gobject_class->set_property = gst_ts_dup_set_property;
  gobject_class->get_property = gst_ts_dup_get_property;

  g_object_class_install_property (gobject_class, PROP_METHOD,
      g_param_spec_enum ("operation", "Operation", "Timestamp Duplicate operation",
          GST_TYPE_TS_DUP_METHOD, DEFAULT_METHOD, G_PARAM_READWRITE));

  gst_element_class_set_static_metadata (element_class, "tsdup",
      "Generic", "Duplicate pts/dts timestamp",
      "Mark Nauwelaerts <mnauw@users.sourceforge.net>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_ts_dup_sink_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_ts_dup_src_template));

  /* not setting passthrough assures (metadata) writable buffers */
  trans_class->transform_ip = GST_DEBUG_FUNCPTR (gst_ts_dup_transform_ip);
  trans_class->start = GST_DEBUG_FUNCPTR (gst_ts_dup_start);
  trans_class->stop = GST_DEBUG_FUNCPTR (gst_ts_dup_stop);
}

static void
gst_ts_dup_init (GstTsDup * filter)
{
  filter->method = DEFAULT_METHOD;
}

static GstFlowReturn
gst_ts_dup_transform_ip (GstBaseTransform * btrans, GstBuffer * in)
{
  GstTsDup *filter = GST_TS_DUP (btrans);
  GstTsDupMethods method = filter->method;

  switch (method) {
    case METHOD_TO_INVALID:
      if (!GST_BUFFER_PTS_IS_VALID (in))
        GST_BUFFER_PTS (in) = GST_BUFFER_DTS (in);
      else
        GST_BUFFER_DTS (in) = GST_BUFFER_PTS (in);
      break;
    case METHOD_TO_PTS:
    case METHOD_TO_INVALID_PTS:
      if (method == METHOD_TO_INVALID_PTS && GST_BUFFER_PTS_IS_VALID (in))
        break;
      GST_BUFFER_PTS (in) = GST_BUFFER_DTS (in);
      break;
    case METHOD_TO_DTS:
    case METHOD_TO_INVALID_DTS:
      if (method == METHOD_TO_INVALID_DTS && GST_BUFFER_DTS_IS_VALID (in))
        break;
      GST_BUFFER_DTS (in) = GST_BUFFER_PTS (in);
      break;
    default:
      break;
  }

  return GST_FLOW_OK;
}


static gboolean
gst_ts_dup_start (GstBaseTransform * btrans)
{
  return TRUE;
}

static gboolean
gst_ts_dup_stop (GstBaseTransform * btrans)
{
  return TRUE;
}

static void
gst_ts_dup_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstTsDup *src;

  g_return_if_fail (GST_IS_TS_DUP (object));
  src = GST_TS_DUP (object);

  switch (prop_id) {
    case PROP_METHOD:
      src->method = g_value_get_enum (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ts_dup_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstTsDup *src;

  g_return_if_fail (GST_IS_TS_DUP (object));
  src = GST_TS_DUP (object);

  switch (prop_id) {
    case PROP_METHOD:
      g_value_set_enum (value, src->method);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}
